#include "VCalendar.hpp"

// Directory name
#define DIR_NAME         "/calendar/"
// Index file name
#define INDEX_FILE_NAME  "index.dat"
// Extension name of VCalendar
#define EXTENSION_NAME   ".vcs"

VCalendar::VCalendar(Void)
{
    return;
}

VCalendar::VCalendar(SFXAnsiStringConstRef string)
{
    Import(string);
    return;
}

VCalendar::VCalendar(SFXAnsiStringConstRef version, SFXDateConstRef startDate,
                     SFXDateConstRef endDate, SFXDateConstRef lastModifiedDate,
                     SFXAnsiStringConstRef summary) : _version(version)
{
    Set(startDate, endDate, lastModifiedDate, summary);
    return;
}

// Set data
Void VCalendar::Set(SFXDateConstRef startDate, SFXDateConstRef endDate,
    SFXDateConstRef lastModifiedDate, SFXAnsiStringConstRef summary)
{
    _startDate = startDate;
    _endDate = endDate;
    _lastModifiedDate = lastModifiedDate;
    _summary = summary;
    return;
}

// Set version
Void VCalendar::SetVersion(SFXAnsiStringConstRef version)
{
    _version = version;
    return;
}

SFXDate VCalendar::GetStartDate(Void) const
{
    return _startDate;
}

SFXDate VCalendar::GetEndDate(Void) const
{
    return _endDate;
}

SFXDate VCalendar::GetLastModifiedDate(Void) const
{
    return _lastModifiedDate;
}

SFXAnsiString VCalendar::GetSummary(Void) const
{
    return _summary;
}

// A character string of vCalendar form is imported. 
Void VCalendar::Import(SFXAnsiStringConstRef string)
{
    SInt32 c1, c2;
    SFXAnsiStringConst format("YYYYMMDD%Thhmmss");
    SFXAnsiString field, value;

    _others.Clear();
    // Error if "BEGIN:VCALENDAR" does not exists at the beginning 
    c1 = string.FirstIndexOf("BEGIN:VCALENDAR");
    if (c1 < 0) {
        c1 = string.FirstIndexOf("begin:vcalendar");
        if (c1 < 0) {
            return;
        }
    }
    c2 = string.FirstIndexOf('\n', c1);
    while (true) {
        // Look for the next ":"
        c1 = string.FirstIndexOf(':', c2); 
        if (c1 < 0) {   // If it is not found
            break;      // End
        }
        // Get the left side of the colon "****:****"; 
        // And cut any blank characters at the end  
        field = string.Substring(c2 + 1, c1).Trim();

        // Search for the next end of line.
        c2 = string.FirstIndexOf('\n', c1); 
        if (c2 < 0) {                   // If it is not found
            c2 = string.GetLength();    // Make it the end of the string 
        }

        // Get the right side of the colon "****:****"; 
        // And cut any blank characters at the end  
        value = string.Substring(c1 + 1, c2).Trim();

        if (field.Equals("begin", false)) { 
            ; // If field is "begin", disregard
        } else if (field.Equals("end", false)){
            if (value.Equals("vcalendar"), false) {
                break; // End when found END:VCALENDAR
            } // Disregard anything else
        } else if (field.Equals("version", false)) { // Version
            _version = value;                        // Set version
        } else if (field.Equals("dtstart", false)) { // Start time
            _startDate.Parse(format, value);         // Parse and set start date 
        } else if (field.Equals("dtend", false)) {   // Finish time
            _endDate.Parse(format, value);           // Parse and set end date  
        } else if (field.Equals("last-modified", false)) { // Update time
            _lastModifiedDate.Parse(format, value);  // Parse and set last-modified date
        // Event data
        } else if (field.Substring(0, 7).Equals("summary", false)) {
            _summary = value;
        } else { // Other data
            _others << field << ":" << value << "\r\n";
        }
    }
    return;
}

// Write the data of the vCalendar
Void VCalendar::Export(SFXAnsiStringPtr string) const
{
    SFXAnsiString format("YYYYMMDD%Thhmmss%Z");

    string->Set("BEGIN:VCALENDAR\r\nVERSION:");
    *string += _version;
    *string += "\r\nBEGIN:VEVENT\r\nCLASS:PUBLIC\r\nDTSTART:";
    *string += _startDate.Format(format);
    *string += "\r\nDTEND:";
    *string += _endDate.Format(format);
    *string += "\r\nSUMMARY;ENCODING=QUOTED-PRINTABLE:";
    *string += _summary;
    *string += "\r\nLAST-MODIFIED:";
    *string += _lastModifiedDate.Format(format);
    *string += "\r\n";
    *string += _others;
    *string += "END:VEVENT\r\nEND:VCALENDAR\r\n";
    return;
}

// Publish the ID to identify data uniquelyi= file name
// ID = YYYYMMDDHHmmssXXX XXX is random number
SFXAnsiString VCalendar::PublishXid(SFXDateConstRef currentDate)
{
    SFXLCGRandom random;
    SFXAnsiString result;

    result = currentDate.Format("YYYYMMDD%Thhmmss");
    result += SFXAnsiString::Format("%03d", random.GetUInt32() % 1000);
    return result;
}

VCalendarFile::VCalendarFile(SFXPathConstRef fileName, SFXAnsiStringConstRef flag)
        : VCalendar(), _fileName(fileName), _flag(flag)
{
    return;
}

VCalendarFile::VCalendarFile(SFXAnsiStringConstRef content,
    SFXPathConstRef fileName, SFXAnsiStringConstRef flag)
        : VCalendar(content), _fileName(fileName), _flag(flag)
{
    return;
}

VCalendarFile::VCalendarFile(SFXAnsiStringConstRef version,
    SFXDateConstRef startDate, SFXDateConstRef endDate,
    SFXAnsiStringConstRef summary) : _flag("-")
{
    SFXDate lastModifiedDate = SFXDate::CurrentDate();
    // The file name is generated from last-modified. 
    _fileName.Set(VCalendar::PublishXid(lastModifiedDate) + EXTENSION_NAME);
    Set(startDate, endDate, lastModifiedDate, summary);
    SetVersion(version);
    return;
}

Void VCalendarFile::SetFileName(SFXPathConstRef fileName)
{
    _fileName = fileName;
    return;
}

Void VCalendarFile::SetFlag(SFXAnsiStringConstRef flag)
{
    _flag = flag;
    return;
}

SFXPath VCalendarFile::GetFileName(Void) const
{
    return _fileName;
}

SFXAnsiString VCalendarFile::GetFlag(Void) const
{
    return _flag;
}

// Data is saved in the file. 
SFCError VCalendarFile::SaveToFile(Void) const
{
    SFCError error;
    SFXFile file;
    SFXAnsiStringStreamWriter writer;
    SFXPath fileName(DIR_NAME + _fileName.Get());
    SFXPath temp;
    SFXAnsiString content;

    // Get temporary file name
    error = SFXFile::GetTemporaryPath(SFXPath(DIR_NAME), &temp);
    if (error == SFERR_NO_ERROR) {
        // Make temporary file
        error = file.OpenReadWrite(temp);
        if (error == SFERR_NO_ERROR) {
            // Convert from VCalendar class into string in vCalendar format
            Export(&content);
            // Get stream for writing
            error = file.GetStreamWriter(content.GetLength(), &writer);
            if (error == SFERR_NO_ERROR) {
                // Write to file
                writer << content;
                writer.Flush();
            }
            // Close file
            file.Close();
        }
    }
    if (error == SFERR_NO_ERROR) {
        // Delete old file
        SFXFile::Remove(fileName);
        // Rename temporary file to original name
        error = SFXFile::Rename(temp, fileName);
    }
    return error;
}

// Read out data from the file. 
SFCError VCalendarFile::LoadFromFile(Void)
{
    SFCError error;
    SFXFile file;
    SFXAnsiStringStreamReader reader;
    UInt32 size;
    SFXAnsiString string;
    SFXPath filename(SFXPath(DIR_NAME + _fileName.Get()));

    // Get file size
    error = SFXFile::GetSize(filename, &size);
    if (error == SFERR_NO_ERROR) {
        // Open file
        error = file.OpenReadOnly(filename);
        if (error == SFERR_NO_ERROR) {
            // Get stream for reading
            error = file.GetStreamReader(size, &reader);
            if (error == SFERR_NO_ERROR) {
                reader.Fetch();
                // Read from file
                reader >> string;
            }
            file.Close();
        }
    }
    if (error == SFERR_NO_ERROR) {
        // Analyze data in vCalendar format
        // Subsutitute each element into member variable 
        Import(string); 
    }
    return error;
}

VCalCollection::VCalCollection(Void) : _lastSyncTime("00000000T000000Z")
{
    return;
}

VCalCollection::~VCalCollection(Void)
{
    SFXList<VCalendarFilePtr>::Enumerator e = _list.GetFirstEnumerator();
    while(e.HasNext()) {
        VCalendarFilePtr vcal = e.GetNext();
        delete vcal;
    }
    return;
}

Void VCalCollection::SetLastSyncTime(SFXAnsiStringConstRef lastSyncTime)
{
    _lastSyncTime = lastSyncTime;
    return;
}

SFXAnsiString VCalCollection::GetLastSyncTime(Void) const
{
    return _lastSyncTime;
}


// Data is added to VCalCollection. 
Void VCalCollection::Append(VCalendarFilePtr vcalendar)
{
    SFXDateConst date1(vcalendar->GetStartDate());
    Bool isInsert = false;
    SIntN i;

    // The flag that indicates a new creation is set. 
    vcalendar->SetFlag("n"); 
    // Search for the position where the element is to be inserted. 
    // Data must be sorted. 
    for (i = 0; i < _list.GetSize(); ++i) {
        // Get the startDate of i th element        
        SFXDate date2 = _list.Get(i)->GetStartDate();
        // Compare the StartDates
        if (date1 <= date2) {   
            // Insert an element
            _list.Insert(i, vcalendar);   
            isInsert = true;
            break;
        }
    }
    // If the insertion position is not found
    if (!isInsert) { 
        _list.Append(vcalendar); // Append it to the end
    }
    // Save data to file 
    vcalendar->SaveToFile(); 
    WriteIndexFile(); // Update index
    return;
}

// Update data
Void VCalCollection::Update(VCalendarFilePtr vcalendar)
{
    _list.Remove(_list.FirstIndexOf(vcalendar));
    Append(vcalendar);
    vcalendar->SetFlag("m");
    return;
}

// The deletion flag "d" is set in data. 
Void VCalCollection::Delete(VCalendarFilePtr vcalendar)
{
    vcalendar->SetFlag("d");
    WriteIndexFile();
    return;
}

// Data is completely removed. 
Void VCalCollection::Remove(VCalendarFilePtr vcalendar)
{
    // Delete file
    SFXFile::Remove(SFXPath(DIR_NAME + vcalendar->GetFileName().Get() + EXTENSION_NAME));
    _list.Remove(_list.FirstIndexOf(vcalendar));
    WriteIndexFile();
    return;
}

// Get enumerator by index
SFXList<VCalendarFilePtr>::Enumerator VCalCollection::GetEnumerator(SInt32 index) const
{
    return _list.GetEnumerator(index);
}

VCalendarFilePtr VCalCollection::Get(SInt32 index) const
{
    return _list.Get(index);
}

VCalendarFilePtr VCalCollection::Get(SFXPathConstRef fileName) const
{
    SFXList<VCalendarFilePtr>::Enumerator e = _list.GetFirstEnumerator();
    while(e.HasNext()) {
        VCalendarFilePtr vcal = e.GetNext();
        if (vcal->GetFileName() == fileName) {
            return vcal;
        }
    }
    return null;
}

// Year, month, day following first date is acquired with index. 
SInt32 VCalCollection::GetIndex(SInt16 year, SInt16 month, SInt16 day) const
{
    SInt32 i;
    SFXDate date1(year, month, day);
    for (i = 0 ; i < _list.GetSize() ; i++) {
        VCalendarFilePtr vcal = _list.Get(i);
        SFXDate date2 = vcal->GetStartDate();
        date2.SetTime(0, 0, 0);
        if (date1 <= date2) {
            break;
        }
    }
    return i;
}

// Get the value of year, month, day
SInt32 VCalCollection::GetNumberOfVCal(SInt32 todayIndex,
                                    SInt16 year, SInt16 month, SInt16 day) const
{
    SFXDate date1(year, month, day);
    SInt32 size = 0;
    SFXList<VCalendarFilePtr>::Enumerator e = _list.GetEnumerator(todayIndex);
    while(e.HasNext()) {
        VCalendarFilePtr vcal = e.GetNext();
        SFXDate date2 = vcal->GetStartDate();
        date2.SetTime(0, 0, 0);
        if (date1 != date2) {
            break;
        }
        ++size; // If the date is the same, +1
    }
    return size;
}

// If the data of year, month, day exists
Bool VCalCollection::IsDataExist(SInt16 year, SInt16 month, SInt16 day) const
{
    SFXList<VCalendarFilePtr>::Enumerator e =
        _list.GetEnumerator(GetIndex(year, month, day));
    if (e.HasNext()) {
        SFXDate date = e.GetNext()->GetStartDate();
        if (date.GetYear() == year && date.GetMonth() == month
        && date.GetDay() == day) {
            return true;
        }
    }
    return false;
}

// All data is read from the file. 
// First , read from the index file. 
// Next, index file data is analyzed, and data is read from each file to VCalCollection. 
SFCError VCalCollection::ReadIndexFile(Void)
{
    SFCError error;
    SFXFile file;
    SFXAnsiStringStreamReader reader;
    UInt32 size;
    SFXAnsiString string;

	//Improved by Yuu
	//if file dosn't exist, Crate it
	bool exists;
	error =SFXFile::Exists(	SFXPath(DIR_NAME INDEX_FILE_NAME),&exists);
	if (exists==false)
	{
		//force to crate a file
		file.Create(SFXPath(DIR_NAME INDEX_FILE_NAME),true);
	}

    // Get file size
    error = SFXFile::GetSize(SFXPath(DIR_NAME INDEX_FILE_NAME), &size);
    if (error == SFERR_NO_ERROR) {
        // Open file
        error = file.OpenReadOnly(SFXPath(DIR_NAME INDEX_FILE_NAME));
        if (error == SFERR_NO_ERROR) {
            // Get stream for reading
            error = file.GetStreamReader(size, &reader);
            if (error == SFERR_NO_ERROR) {
                reader.Fetch();
                // Read from file
                reader >> string;
            }
            file.Close();
        }
    }
    if (error == SFERR_NO_ERROR) {
        SInt32 c1 = 0;
        SInt32 c2 = 0;
		// Search for end of line
        c2 = string.FirstIndexOf('\n');
		// Trim blank characters
        _lastSyncTime = string.Substring(0, c2).Trim();
        ++c2;
        while ((c1 = string.FirstIndexOf(' ', c2)) > -1) {
            SFXPath filename(string.Substring(c2, c1));
            c2 = string.FirstIndexOf("\r\n", c1);
            SFXAnsiString flag(string.Substring(c1 + 1, c2));
            c2 += 2; 

            VCalendarFilePtr vcal = ::new VCalendarFile(filename, flag);
            // The vCalendar data is read from the file. 
            vcal->LoadFromFile(); 
            _list.Append(vcal);
        }
    }
    return error;
}

// Write file
SFCError VCalCollection::WriteIndexFile(Void)
{
    SFCError error;
    SFXFile file;
    SFXAnsiStringStreamWriter writer;
    SFXPath fileName(DIR_NAME INDEX_FILE_NAME);
    SFXPath temp;
    SFXAnsiString content;

    // Get temporary file name
    error = SFXFile::GetTemporaryPath(SFXPath(DIR_NAME), &temp);
    if (error == SFERR_NO_ERROR) {
        // Make temporarily file
        error = file.OpenReadWrite(temp);
        if (error == SFERR_NO_ERROR) {
            // Get stream for writing
            error = file.GetStreamWriter(1024, &writer);
            if (error == SFERR_NO_ERROR) {
                // Write to file
                writer << _lastSyncTime << "\r\n";
                writer.Flush();
                SFXList<VCalendarFilePtr>::Enumerator etor
                    = _list.GetFirstEnumerator();
                while(etor.HasNext()) {
                    VCalendarFilePtr vcal = etor.GetNext();
                    writer << vcal->GetFileName().Get() << " "
                        << vcal->GetFlag() << "\r\n";
                    writer.Flush();
                }
            }
            // Close file
            file.Close();
        }
    }
    if (error == SFERR_NO_ERROR) {
        // Delete old file
        SFXFile::Remove(fileName);
        // Rename temporary file to original name
        error = SFXFile::Rename(temp, fileName);
    }
    return error;
}

SFXAnsiString VCalCollection::MakeListForSync(Void)
{
    SFXAnsiString listStr(_lastSyncTime);
    SFXList<VCalendarFilePtr>::Enumerator e = _list.GetFirstEnumerator();
    while (e.HasNext()) {
        VCalendarFilePtr vcal = e.GetNext();
        if (vcal->GetFlag() != "-") {
            SFXAnsiString content;
            vcal->Export(&content);
            listStr << "####" << vcal->GetFlag() << "#!#!"
                << vcal->GetFileName().Get() << "#!#!" << content << "\r\n";
        }
    }
    return listStr;
}
